=== modified file 'bzrlib/hashcache.py'
--- bzrlib/hashcache.py	2006-03-20 18:34:33 +0000
+++ bzrlib/hashcache.py	2006-07-12 05:22:48 +0000
@@ -42,20 +42,6 @@
 FP_CTIME_COLUMN = 2
 FP_MODE_COLUMN = 5
 
-def _fingerprint(abspath):
-    try:
-        fs = os.lstat(abspath)
-    except OSError:
-        # might be missing, etc
-        return None
-
-    if stat.S_ISDIR(fs.st_mode):
-        return None
-
-    # we discard any high precision because it's not reliable; perhaps we
-    # could do better on some systems?
-    return (fs.st_size, long(fs.st_mtime),
-            long(fs.st_ctime), fs.st_ino, fs.st_dev, fs.st_mode)
 
 
 class HashCache(object):
@@ -131,7 +117,7 @@
         
         for inum, path, cache_entry in prep:
             abspath = pathjoin(self.root, path)
-            fp = _fingerprint(abspath)
+            fp = self._fingerprint(abspath)
             self.stat_count += 1
             
             cache_fp = cache_entry[1]
@@ -142,13 +128,12 @@
                 self.needs_write = True
                 del self._cache[path]
 
-
     def get_sha1(self, path):
         """Return the sha1 of a file.
         """
         abspath = pathjoin(self.root, path)
         self.stat_count += 1
-        file_fp = _fingerprint(abspath)
+        file_fp = self._fingerprint(abspath)
         
         if not file_fp:
             # not a regular file or not existing
@@ -164,22 +149,26 @@
             cache_sha1, cache_fp = None, None
 
         if cache_fp == file_fp:
+            ## mutter("hashcache hit for %s %r -> %s", path, file_fp, cache_sha1)
+            ## mutter("now = %s", time.time())
             self.hit_count += 1
             return cache_sha1
         
         self.miss_count += 1
 
-
         mode = file_fp[FP_MODE_COLUMN]
         if stat.S_ISREG(mode):
-            digest = sha_file(file(abspath, 'rb', buffering=65000))
+            digest = self._really_sha1_file(abspath)
         elif stat.S_ISLNK(mode):
             digest = sha.new(os.readlink(abspath)).hexdigest()
         else:
             raise BzrError("file %r: unknown file stat mode: %o"%(abspath,mode))
 
-        now = int(time.time())
-        if file_fp[FP_MTIME_COLUMN] >= now or file_fp[FP_CTIME_COLUMN] >= now:
+        # window of 3 seconds to allow for 2s resolution on windows,
+        # unsynchronized file servers, etc.
+        cutoff = self._cutoff_time()
+        if file_fp[FP_MTIME_COLUMN] >= cutoff \
+                or file_fp[FP_CTIME_COLUMN] >= cutoff:
             # changed too recently; can't be cached.  we can
             # return the result and it could possibly be cached
             # next time.
@@ -191,16 +180,24 @@
             # need to let sufficient time elapse before we may cache this entry
             # again.  If we didn't do this, then, for example, a very quick 1
             # byte replacement in the file might go undetected.
-            self.danger_count += 1 
+            ## mutter('%r modified too recently; not caching', path)
+            self.danger_count += 1
             if cache_fp:
                 self.removed_count += 1
                 self.needs_write = True
                 del self._cache[path]
         else:
+            ## mutter('%r added to cache: now=%f, mtime=%d, ctime=%d',
+            ##        path, time.time(), file_fp[FP_MTIME_COLUMN],
+            ##        file_fp[FP_CTIME_COLUMN])
             self.update_count += 1
             self.needs_write = True
             self._cache[path] = (digest, file_fp)
         return digest
+
+    def _really_sha1_file(self, abspath):
+        """Calculate the SHA1 of a file by reading the full text"""
+        return sha_file(file(abspath, 'rb', buffering=65000))
         
     def write(self):
         """Write contents of cache to file."""
@@ -218,10 +215,10 @@
                 print >>outf
             outf.commit()
             self.needs_write = False
-            mutter("write hash cache: %s hits=%d misses=%d stat=%d recent=%d updates=%d",
-                   self.cache_file_name(), self.hit_count, self.miss_count,
-                   self.stat_count,
-                   self.danger_count, self.update_count)
+            ## mutter("write hash cache: %s hits=%d misses=%d stat=%d recent=%d updates=%d",
+            ##        self.cache_file_name(), self.hit_count, self.miss_count,
+            ##        self.stat_count,
+            ##        self.danger_count, self.update_count)
         finally:
             if not outf.closed:
                 outf.abort()
@@ -244,7 +241,6 @@
             self.needs_write = True
             return
 
-
         hdr = inf.readline()
         if hdr != CACHE_HEADER:
             mutter('cache header marker not found at top of %s;'
@@ -275,7 +271,24 @@
             self._cache[path] = (sha1, fp)
 
         self.needs_write = False
+
+    def _cutoff_time(self):
+        """Return cutoff time.
+
+        Files modified more recently than this time are at risk of being
+        undetectably modified and so can't be cached.
+        """
+        return int(time.time()) - 3
            
-
-
-        
+    def _fingerprint(self, abspath):
+        try:
+            fs = os.lstat(abspath)
+        except OSError:
+            # might be missing, etc
+            return None
+        if stat.S_ISDIR(fs.st_mode):
+            return None
+        # we discard any high precision because it's not reliable; perhaps we
+        # could do better on some systems?
+        return (fs.st_size, long(fs.st_mtime),
+                long(fs.st_ctime), fs.st_ino, fs.st_dev, fs.st_mode)

=== modified file 'bzrlib/tests/test_hashcache.py'
--- bzrlib/tests/test_hashcache.py	2006-03-31 00:43:21 +0000
+++ bzrlib/tests/test_hashcache.py	2006-07-12 05:22:48 +0000
@@ -1,4 +1,4 @@
-# (C) 2005 Canonical Ltd
+# Copyright (C) 2005, 2006 Canonical Ltd
 
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -16,12 +16,13 @@
 
 import os
 import sha
+import stat
 import sys
 import time
 
 from bzrlib.errors import BzrError
 from bzrlib.hashcache import HashCache
-from bzrlib.tests import TestCaseInTempDir, TestSkipped
+from bzrlib.tests import TestCaseInTempDir, TestSkipped, TestCase
 
 
 def sha1(t):
@@ -29,117 +30,180 @@
 
 
 def pause():
-    if False:
-        return
-    if sys.platform in ('win32', 'cygwin'):
-        time.sleep(3)
-        return
-    # allow it to stabilize
-    start = int(time.time())
-    while int(time.time()) == start:
-        time.sleep(0.2)
-
-
-class FixThisError(Exception):
-    pass
-    
+    time.sleep(5.0)
+
 
 class TestHashCache(TestCaseInTempDir):
-
-    def test_hashcache(self):
-        """Functional tests for hashcache"""
-
+    """Test the hashcache against a real directory"""
+
+    def make_hashcache(self):
         # make a dummy bzr directory just to hold the cache
         os.mkdir('.bzr')
         hc = HashCache('.', '.bzr/stat-cache')
-
-        file('foo', 'wb').write('hello')
-        os.mkdir('subdir')
-        pause()
-
+        return hc
+
+    def reopen_hashcache(self):
+        hc = HashCache('.', '.bzr/stat-cache')
+        hc.read()
+        return hc
+
+    def test_hashcache_initial_miss(self):
+        """Get correct hash from an empty hashcache"""
+        hc = self.make_hashcache()
+        self.build_tree_contents([('foo', 'hello')])
         self.assertEquals(hc.get_sha1('foo'),
                           'aaf4c61ddcc5e8a2dabede0f3b482cd9aea9434d')
         self.assertEquals(hc.miss_count, 1)
         self.assertEquals(hc.hit_count, 0)
 
-        # check we hit without re-reading
-        self.assertEquals(hc.get_sha1('foo'),
-                          'aaf4c61ddcc5e8a2dabede0f3b482cd9aea9434d')
-        self.assertEquals(hc.miss_count, 1)
-        self.assertEquals(hc.hit_count, 1)
-
-        # check again without re-reading
-        self.assertEquals(hc.get_sha1('foo'),
-                          'aaf4c61ddcc5e8a2dabede0f3b482cd9aea9434d')
-        self.assertEquals(hc.miss_count, 1)
-        self.assertEquals(hc.hit_count, 2)
-
-        # write new file and make sure it is seen
-        file('foo', 'wb').write('goodbye')
-        pause()
-        self.assertEquals(hc.get_sha1('foo'),
-                          '3c8ec4874488f6090a157b014ce3397ca8e06d4f')
-        self.assertEquals(hc.miss_count, 2)
-
-        # quickly write new file of same size and make sure it is seen
-        # this may rely on detection of timestamps that are too close
-        # together to be safe
-        file('foo', 'wb').write('g00dbye')
-        self.assertEquals(hc.get_sha1('foo'),
-                          sha1('g00dbye'))
-
-        file('foo2', 'wb').write('other file')
-        self.assertEquals(hc.get_sha1('foo2'), sha1('other file'))
-
-        os.remove('foo2')
-        self.assertEquals(hc.get_sha1('foo2'), None)
-
-        file('foo2', 'wb').write('new content')
-        self.assertEquals(hc.get_sha1('foo2'), sha1('new content'))
-
+    def test_hashcache_new_file(self):
+        hc = self.make_hashcache()
+        self.build_tree_contents([('foo', 'goodbye')])
+        # now read without pausing; it may not be possible to cache it as its
+        # so new
+        self.assertEquals(hc.get_sha1('foo'), sha1('goodbye'))
+
+    def test_hashcache_nonexistent_file(self):
+        hc = self.make_hashcache()
+        self.assertEquals(hc.get_sha1('no-name-yet'), None)
+
+    def test_hashcache_replaced_file(self):
+        hc = self.make_hashcache()
+        self.build_tree_contents([('foo', 'goodbye')])
+        self.assertEquals(hc.get_sha1('foo'), sha1('goodbye'))
+        os.remove('foo')
+        self.assertEquals(hc.get_sha1('foo'), None)
+        self.build_tree_contents([('foo', 'new content')])
+        self.assertEquals(hc.get_sha1('foo'), sha1('new content'))
+
+    def test_hashcache_not_file(self):
+        hc = self.make_hashcache()
+        self.build_tree(['subdir/'])
         self.assertEquals(hc.get_sha1('subdir'), None)
 
-        # pause briefly to make sure they're not treated as new uncacheable
-        # files
+    def test_hashcache_load(self):
+        hc = self.make_hashcache()
+        self.build_tree_contents([('foo', 'contents')])
         pause()
-
-        self.assertEquals(hc.get_sha1('foo'), sha1('g00dbye'))
-        self.assertEquals(hc.get_sha1('foo2'), sha1('new content'))
-
-        # write out, read back in and check that we don't need to
-        # re-read any files
+        self.assertEquals(hc.get_sha1('foo'), sha1('contents'))
         hc.write()
-        del hc
-
-        hc = HashCache('.', '.bzr/stat-cache')
-        hc.read()
-
-        self.assertEquals(len(hc._cache), 2)
-        self.assertEquals(hc.get_sha1('foo'), sha1('g00dbye'))
+        hc = self.reopen_hashcache()
+        self.assertEquals(hc.get_sha1('foo'), sha1('contents'))
         self.assertEquals(hc.hit_count, 1)
-        self.assertEquals(hc.miss_count, 0)
-        self.assertEquals(hc.get_sha1('foo2'), sha1('new content'))
+
+    def test_hammer_hashcache(self):
+        hc = self.make_hashcache()
+        for i in xrange(10000):
+            self.log('start writing at %s', time.time())
+            f = file('foo', 'w')
+            try:
+                last_content = '%08x' % i
+                f.write(last_content)
+            finally:
+                f.close()
+            last_sha1 = sha1(last_content)
+            self.log("iteration %d: %r -> %r",
+                     i, last_content, last_sha1)
+            got_sha1 = hc.get_sha1('foo')
+            self.assertEquals(got_sha1, last_sha1)
+            hc.write()
+            hc = self.reopen_hashcache()
 
     def test_hashcache_raise(self):
         """check that hashcache can raise BzrError"""
-
-        os.mkdir('.bzr')
-        hc = HashCache('.', '.bzr/stat-cache')
-        ok = False
-
-        # make a best effort to create a weird kind of file
-        funcs = (getattr(os, 'mkfifo', None), getattr(os, 'mknod', None))
-        for func in funcs:
-            if func is None:
-                continue
-            try:
-                func('a')
-                ok = True
-                break
-            except FixThisError:
-                pass
-
-        if ok:
-            self.assertRaises(BzrError, hc.get_sha1, 'a')
+        hc = self.make_hashcache()
+        if not hasattr(os, 'mkfifo'):
+            raise TestSkipped('filesystem fifos not supported on this system')
+        os.mkfifo('a')
+        # It's possible that the system supports fifos but the filesystem
+        # can't.  In that case we should skip at this point.  But in fact
+        # such combinations don't usually occur for the filesystem where
+        # people test bzr.
+        self.assertRaises(BzrError, hc.get_sha1, 'a')
+
+
+class FakeHashCache(HashCache):
+    """Hashcache that consults a fake clock rather than the real one.
+
+    This lets us examine how old or new files would be handled, without
+    actually having to wait for time to pass.
+    """
+    def __init__(self):
+        # set root and cache file name to none to make sure we won't touch the
+        # real filesystem
+        HashCache.__init__(self, '.', 'hashcache')
+        self._files = {}
+        # simulated clock running forward as operations happen
+        self._clock = 0
+
+    def put_file(self, filename, file_contents):
+        abspath = './' + filename
+        self._files[abspath] = (file_contents, self._clock)
+
+    def _fingerprint(self, abspath):
+        entry = self._files[abspath]
+        return (len(entry[0]),
+                entry[1], entry[1],
+                10, 20,
+                stat.S_IFREG | 0600)
+
+    def _really_sha1_file(self, abspath):
+        if abspath in self._files:
+            return sha1(self._files[abspath][0])
         else:
-            raise TestSkipped('No weird file type could be created')
+            return None
+
+    def _cutoff_time(self):
+        return self._clock - 2
+
+    def pretend_to_sleep(self, secs):
+        self._clock += secs
+
+    
+class TestHashCacheFakeFilesystem(TestCaseInTempDir):
+    """Tests the hashcache using a simulated OS.
+    """
+
+    def make_hashcache(self):
+        return FakeHashCache()
+
+    def test_hashcache_miss_new_file(self):
+        """A new file gives the right sha1 but misses"""
+        hc = self.make_hashcache()
+        hc.put_file('foo', 'hello')
+        self.assertEquals(hc.get_sha1('foo'), sha1('hello'))
+        self.assertEquals(hc.miss_count, 1)
+        self.assertEquals(hc.hit_count, 0)
+        # if we try again it's still too new; 
+        self.assertEquals(hc.get_sha1('foo'), sha1('hello'))
+        self.assertEquals(hc.miss_count, 2)
+        self.assertEquals(hc.hit_count, 0)
+
+    def test_hashcache_old_file(self):
+        """An old file gives the right sha1 and hits"""
+        hc = self.make_hashcache()
+        hc.put_file('foo', 'hello')
+        hc.pretend_to_sleep(20)
+        # file is new; should get the correct hash but miss
+        self.assertEquals(hc.get_sha1('foo'), sha1('hello'))
+        self.assertEquals(hc.miss_count, 1)
+        self.assertEquals(hc.hit_count, 0)
+        # and can now be hit
+        self.assertEquals(hc.get_sha1('foo'), sha1('hello'))
+        self.assertEquals(hc.miss_count, 1)
+        self.assertEquals(hc.hit_count, 1)
+        hc.pretend_to_sleep(3)
+        # and again
+        self.assertEquals(hc.get_sha1('foo'), sha1('hello'))
+        self.assertEquals(hc.miss_count, 1)
+        self.assertEquals(hc.hit_count, 2)
+
+    def test_hashcache_invalidates(self):
+        hc = self.make_hashcache()
+        hc.put_file('foo', 'hello')
+        hc.pretend_to_sleep(20)
+        hc.get_sha1('foo')
+        hc.put_file('foo', 'h1llo')
+        self.assertEquals(hc.get_sha1('foo'), sha1('h1llo'))
+        self.assertEquals(hc.miss_count, 2)
+        self.assertEquals(hc.hit_count, 0)

=== modified file 'bzrlib/tests/workingtree_implementations/test_workingtree.py'
--- bzrlib/tests/workingtree_implementations/test_workingtree.py	2006-07-11 18:34:09 +0000
+++ bzrlib/tests/workingtree_implementations/test_workingtree.py	2006-07-12 10:09:27 +0000
@@ -182,20 +182,6 @@
         self.assertEquals(list(tree.unknowns()),
                           ['hello.txt'])
 
-    def test_hashcache(self):
-        from bzrlib.tests.test_hashcache import pause
-        tree = self.make_branch_and_tree('.')
-        self.build_tree(['hello.txt',
-                         'hello.txt.~1~'])
-        tree.add('hello.txt')
-        pause()
-        sha = tree.get_file_sha1(tree.path2id('hello.txt'))
-        self.assertEqual(1, tree._hashcache.miss_count)
-        tree2 = WorkingTree.open('.')
-        sha2 = tree2.get_file_sha1(tree2.path2id('hello.txt'))
-        self.assertEqual(0, tree2._hashcache.miss_count)
-        self.assertEqual(1, tree2._hashcache.hit_count)
-
     def test_initialize(self):
         # initialize should create a working tree and branch in an existing dir
         t = self.make_branch_and_tree('.')

